package org.apache.cordova.camera;

import android.app.Activity;
import android.content.ContentResolver;
import android.content.res.Resources;
import android.graphics.Bitmap;
import android.graphics.Color;
import android.hardware.Camera;
import android.net.Uri;
import android.os.Bundle;
import android.provider.MediaStore;
import android.view.View;
import android.widget.FrameLayout;
import android.widget.ImageButton;
import android.widget.TextView;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import org.apache.cordova.LOG;


public class CameraActivity extends Activity {

    private static final String TIME_FORMAT = "yyyyMMdd_HHmmss";

    private final String TAG = "CameraActivity";
    private ContentResolver mContentResolver;
    private Uri mSaveUri;
    private String mSaveFilePath;

    private Camera mCamera;
    private FrameLayout mFrameLayout;
    private CameraPreview mPreview;
    private TextView mText;
    private ImageButton mFlashButton;
    private int mCameraId;

    private String appResourcePackage;

    private final String[] flashModes = {Camera.Parameters.FLASH_MODE_ON, Camera.Parameters.FLASH_MODE_OFF};
    private int fmi = 1; // flash mode index

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        mContentResolver = getContentResolver();

        appResourcePackage = this.getPackageName();

        setContentView(getResources().getIdentifier("activity_camera", "layout", appResourcePackage));

        int cameraPreviewId = getResources().getIdentifier("camera_preview", "id", appResourcePackage);
        int textOverlayId = getResources().getIdentifier("textOverlay", "id", appResourcePackage);
        int buttonFlashId = getResources().getIdentifier("btn_flash", "id", appResourcePackage);

        mFrameLayout = findViewById(cameraPreviewId);
        mText = findViewById(textOverlayId);
        mFlashButton = findViewById(buttonFlashId);

        int width = Resources.getSystem().getDisplayMetrics().widthPixels;
        int height = Resources.getSystem().getDisplayMetrics().heightPixels;
        mText.setText("Place card vertically below within rectangular bracket and take photo");
        mText.setTranslationY(100);
        mText.bringToFront();

        int backCameraID = Util.getBackCameraId();
        mCameraId = backCameraID;

        mCamera = Camera.open(backCameraID);

        mPreview = new CameraPreview(this, mCamera, backCameraID, this);
        mPreview.setBackgroundColor(Color.parseColor("#7f000000"));
        mFrameLayout.addView(mPreview);

        mFlashButton.setOnClickListener(
                new View.OnClickListener() {
                    @Override
                    public void onClick(View v) {
                        setFlashMode();
                    }
                }
        );

    }

    private int getTextPosition(int width, int height) {
        float rectWidth = width * 3.f / 4.f;
        float rectHeight = rectWidth * 9.f / 16.f;
        float margin = (height - rectHeight) / 2.f;

        return (int)(margin - 60);
    }

    private void setFlashMode() {
        Camera.Parameters params = mCamera.getParameters();
        int flashOnId = getResources().getIdentifier("flash_on", "drawable", appResourcePackage);
        int flashOffId = getResources().getIdentifier("flash_off", "drawable", appResourcePackage);
        if (fmi == 1) {
            // flash off
            fmi = 0;
            mFlashButton.setImageResource(flashOnId);
        } else {
            fmi = 1;
            mFlashButton.setImageResource(flashOffId);
        }

        params.setFlashMode(flashModes[fmi]);
        mCamera.setParameters(params);
    }

    Camera.PictureCallback mPictureCallback = new Camera.PictureCallback() {
        @Override
        public void onPictureTaken(byte[] data, Camera camera) {
            Bundle myExtras = getIntent().getExtras();
            if (myExtras != null) {
                mSaveUri = (Uri) myExtras.getParcelable(MediaStore.EXTRA_OUTPUT);
                mSaveFilePath = myExtras.getString("fileName");
            }
            if (mSaveUri != null && mSaveFilePath != null) {
                OutputStream outputStream = null;
                try {
                    int rotation = getPictureRotation(mCameraId);
                    ExifHelper exif = getOriginExifFromDataArray(data);
                    Bitmap bitmap = Util.getBitMapfromByte(data);
                    Bitmap cropmap = Util.centerCrop(bitmap, rotation);
                    byte[] cropdata = Util.getBytefromBitMap(cropmap);
                    outputStream = mContentResolver.openOutputStream(mSaveUri);
                    outputStream.write(cropdata);
                    outputStream.close();

                    exif.createOutFile(mSaveFilePath);
                    exif.writeExifData();
                    setResult(RESULT_OK);
                    finish();
                } catch (IOException e) {
                    // ignore exception
                    e.printStackTrace();
                } finally {
                    Util.closeSilently(outputStream);
                }
            }
        }
    };

    @Override
    protected void onResume() {
        super.onResume();

        if (mCamera == null) {
            mCamera = Camera.open();
        }
    }

    @Override
    protected void onPause() {
        super.onPause();
        releaseCamera();
    }

    public void captureImage(View v) {
        if (mCamera != null) {
            mCamera.autoFocus(new Camera.AutoFocusCallback() {
                @Override
                public void onAutoFocus(boolean success, Camera camera) {
                    if (success) {
                        mCamera.takePicture(null, null, mPictureCallback);
                    }
                }
            });
        }
    }

    private void releaseCamera() {
        if (mCamera != null) {
            mCamera.stopPreview();
            mCamera.setPreviewCallback(null);
            mCamera.release();
            mCamera = null;
        }
    }

    private int getPictureRotation(int cameraId) {
        int deviceOrientation = Util.getDisplayOrientation(this);
        int cameraOrientation = Util.getCameraOrientation(cameraId);
        return Util.getCameraDisplayOrientation(deviceOrientation, cameraOrientation);
    }

    /*

     */
    private ExifHelper getOriginExifFromDataArray(byte[] data) throws IOException {
        ExifHelper exif = null;
        InputStream fileStream = null;
        try {
            fileStream = new ByteArrayInputStream(data);
            if (fileStream != null) {
                exif = new ExifHelper();
                exif.createInFile(fileStream);
                exif.readExifData();
            }
        } catch (IOException e) {
            LOG.d(TAG, "Unable to read exif data: " + e.toString());
        } finally {
            Util.closeSilently(fileStream);
        }

        return exif;
    }

}
